# HSLU
#
# Created by Thomas Koller on 7/28/2020
#
import logging
import numpy as np
from jass.agents.agent import Agent
from jass.game.const import PUSH, MAX_TRUMP, card_strings
from jass.game.game_observation import GameObservation
from jass.game.rule_schieber import RuleSchieber


class AgentRandomSchieber (Agent):
    """
    Randomly select actions for the match of jass (Schieber)
    """
    def __init__(self):
        # log actions
        self._logger = logging.getLogger(__name__)
        # Use rule object to determine valid actions
        self._rule = RuleSchieber()
        # init random number generator
        self._rng = np.random.default_rng()

    def action_trump(self, obs: GameObservation) -> int:
        """
        Select trump randomly. Pushing is selected with probability 0.5 if possible.
        Args:
            obs: the current match
        Returns:
            trump action
        """
        if obs.forehand == -1:
            # if forehand is not yet set, we are the forehand player and can select trump or push
            if self._rng.choice([True, False]):
                return PUSH
        # if not push or forehand, select a trump
        return int(self._rng.integers(low=0, high=MAX_TRUMP, endpoint=True))

    def action_play_card(self, obs: GameObservation) -> int:
        """
        Select randomly a card from the valid cards
        Args:
            obs: The observation of the jass match for the current player
        Returns:
            card to play
        """
        # cards are one hot encoded
        valid_cards = self._rule.get_valid_cards_from_obs(obs)
        # convert to list and draw a value
        card = self._rng.choice(np.flatnonzero(valid_cards))
        self._logger.debug('Played card: {}'.format(card_strings[card]))
        return card




# HSLU
#
# Created by Thomas Koller on 7/28/2020
#
from jass.game.game_observation import GameObservation


class Agent:
    """
    Agent to act as a player in a match of jass.
    """
    def action_trump(self, obs: GameObservation) -> int:
        """
        Determine trump action for the given observation
        Args:
            obs: the match observation, it must be in a state for trump selection

        Returns:
            selected trump as encoded in jass.match.const or jass.match.const.PUSH
        """
        raise NotImplementedError

    def action_play_card(self, obs: GameObservation) -> int:
        """
        Determine the card to play.

        Args:
            obs: the match observation

        Returns:
            the card to play, int encoded as defined in jass.match.const
        """
        raise NotImplementedError




# HSLU
#
# Created by Thomas Koller on 7/23/2020
#
import logging
import numpy as np

from jass.game.const import JASS_SCHIEBER, partner_player, next_player, card_ids
from jass.game.game_util import convert_int_encoded_cards_to_str_encoded, \
    convert_one_hot_encoded_cards_to_str_encoded_list, convert_str_encoded_cards_to_int_encoded


class GameState:
    """
    State of the match.

     A GameState object captures the information in the following stages of the match:
    - Cards have been dealt, but no trump is selected yet
    - The first player that is allowed to choose trump has passed this right to the partner (optional)
    - Trump has been declared by either player from the team that declares trump, but no card has been played yet
    - Between 1 and 35 cards have been played
    - The last card has been played, which is the end of the match.

    The class captures only the data without any logic how to change the data consistently.
    """

    # version of match state
    FORMAT_VERSION = 'V0.2'
    
    def __init__(self) -> None:
        """
        Initialize the class. All numpy arrays will be allocated.
        """
        # dealer of the match
        self.dealer: int = -1

        # player of the next action, i.e. declaring trump or playing a card
        self.player: int = -1

        # selected trump
        self.trump: int = -1

        # true (1) if trump was declared forehand, 0 if it was declared rearhand, -1 if it has not been declared yet
        self.forehand: int = -1

        # the player, who declared trump (derived)
        self.declared_trump: int = -1

        #
        # information about held and played cards
        #

        # the current hands of all the players, 1-hot encoded
        self.hands = np.zeros(shape=[4, 36], dtype=np.int32)

        # the tricks played so far, with the cards of the tricks int encoded in the order they are played
        # a value of -1 indicates that the card has not been played yet
        self.tricks = np.full(shape=[9, 4], fill_value=-1, dtype=np.int32)

        # the winner of the tricks
        self.trick_winner = np.full(shape=9, fill_value=-1, dtype=np.int32)

        # the points made in the tricks
        self.trick_points = np.zeros(shape=9, dtype=np.int32)

        # the first player of the trick (derived)
        self.trick_first_player = np.full(shape=9, fill_value=-1, dtype=np.int32)

        # the current trick is a view onto self.trick
        self.current_trick = self.tricks[0, :]

        # the number of completed tricks
        self.nr_tricks = 0

        # the number of card in the current trick
        self.nr_cards_in_trick = 0

        # the total number of played cards (derived)
        self.nr_played_cards = 0

        self.points = np.zeros(shape=2, dtype=np.int32)

    def __eq__(self, other: 'GameState') -> bool:
        if self.nr_played_cards == 36:
            assert self.current_trick is None
            assert other.current_trick is None
            current_tricks_same = True
        else:
            current_tricks_same = (self.current_trick == other.current_trick).all()
        return self.dealer == other.dealer and \
               self.player == other.player and \
               self.trump == other.trump and \
               self.forehand == other.forehand and \
               self.declared_trump == other.declared_trump and \
               (self.hands == other.hands).all() and \
               (self.tricks == other.tricks).all() and \
               (self.trick_first_player == other.trick_first_player).all() and \
               (self.trick_winner == other.trick_winner).all() and \
               (self.trick_points == other.trick_points).all() and \
               self.nr_tricks == other.nr_tricks and \
               current_tricks_same and \
               self.nr_cards_in_trick == other.nr_cards_in_trick and \
               self.nr_played_cards == other.nr_played_cards and \
               (self.points == other.points).all()

    def __repr__(self):
        return str(self.__dict__)

    def to_json(self):
        """
        Generate a dict representation that can be converted to json. The format is the same than used in
        GameObservation.

        Returns:
            dict representation of the state
        """
        data = dict()

        # change of version: trump and player are always written, even as -1 if they are not known yet
        data['version'] = GameState.FORMAT_VERSION
        data['trump'] = int(self.trump)
        data['dealer'] = int(self.dealer)
        data['currentPlayer'] = int(self.player)
        data['forehand'] = int(self.forehand)

        # played tricks
        tricks = []

        # tricks
        for i in range(min(self.nr_tricks + 1, 9)):
            # add at most one more trick as finished (for first player)
            trick = {}
            # add cards (if not empty)
            cards = convert_int_encoded_cards_to_str_encoded(self.tricks[i, :].tolist())
            if len(cards) > 0:
                trick['cards'] = cards
                if len(cards) == 4:
                    # only add points and winners for complete tricks
                    trick['points'] = int(self.trick_points[i])
                    trick['win'] = int(self.trick_winner[i])

            if self.trick_first_player[i] != -1:
                trick['first'] = int(self.trick_first_player[i])
            if trick:
                # add if not empty dict
                tricks.append(trick)

        data['tricks'] = tricks

        # cards in the hand of the players (empty if no more cards)
        hand_empty = dict(hand=[])
        player_data = [hand_empty, hand_empty, hand_empty, hand_empty]
        for player in range(4):
            hand = dict(hand=convert_one_hot_encoded_cards_to_str_encoded_list(self.hands[player]))
            player_data[player] = hand
            data['player'] = player_data

        data['jassTyp'] = JASS_SCHIEBER
        return data

    def get_card_played(self, card_nr: int) -> int:
        """
        Get the card played at a certain time. Utility method.

        Returns:
            the card that was played as the card_nr card in the match
        """
        nr_trick, card_in_trick = divmod(card_nr, 4)
        return int(self.tricks[nr_trick, card_in_trick])


    @classmethod
    def from_json(cls, data: dict):
        """
        Create state from dict (read from json)
        """
        state = GameState()
        
        # if the version is present, it must be the correct version
        # if it is not there we accept the data for backward compatibility and in the future will assume this
        # version number if absent
        if 'version' in data:
            if data['version'] != GameState.FORMAT_VERSION:
                logging.getLogger(__name__).error('Unexpected format version: {}'.format(data['version']))
                return None

        # not currently used
        # jass_typ = data['jassTyp']

        state.dealer = data['dealer']

        if 'currentPlayer' in data:
            state.player = data['currentPlayer']
        else:
            state.player = -1

        if 'trump' in data:
            state.trump = data['trump']
        else:
            state.trump = -1

        # version 0.2 handling
        if 'forehand' in data:
            state.forehand = data['forehand']
        # previous version handling
        else:
            if 'tss' in data and data['tss'] == 1:
                state.forehand = 0
            elif state.trump != -1:
                # only set if trump has been declared
                state.forehand = 1
            else:
                # beginning of the match, when trump has not been set yet
                state.forehand = -1

        if state.trump != -1:
            if state.forehand == 1:
                state.declared_trump = next_player[state.dealer]
            else:
                state.declared_trump = partner_player[next_player[state.dealer]]

        tricks = data['tricks']
        for i, trick in enumerate(tricks):
            # list of strings in the trick
            if 'cards' in trick:
                cards = trick['cards']
                state.nr_played_cards += len(cards)

                # convert to list of ids
                cards_in_trick = convert_str_encoded_cards_to_int_encoded(cards)
                # append -1 if trick is not complete
                while len(cards_in_trick) < 4:
                    cards_in_trick.append(-1)

                state.tricks[i] = np.array(cards_in_trick)
            if 'win' in trick:
                state.trick_winner[i] = trick['win']
            if 'points' in trick:
                state.trick_points[i] = trick['points']
            # first must be present for all tricks
            if 'first' in trick:
                state.trick_first_player[i] = trick['first']
            else:
                logging.getLogger(__name__).error('No first player set in trick {}'.format(i))

        state.nr_tricks, state.nr_cards_in_trick = divmod(state.nr_played_cards, 4)

        # current trick points to the correct trick
        if state.nr_played_cards != 36:
            state.current_trick = state.tricks[state.nr_tricks]
        else:
            state.current_trick = None

        for i, player_data in enumerate(data['player']):
            if 'hand' in player_data and len(player_data['hand']) > 0:
                hand = player_data['hand']
                for card_constant in hand:
                    state.hands[i, card_ids[card_constant]] = 1

        state.points[0] = 0
        state.points[1] = 0
        for trick in range(state.nr_tricks):
            if state.trick_winner[trick] == 0 or state.trick_winner[trick] == 2:
                state.points[0] += state.trick_points[trick]
            else:
                state.points[1] += state.trick_points[trick]
        return state






# HSLU
#
# Created by Thomas Koller on 7/25/2020
#
import numpy as np

from typing import List, Iterable

from jass.game.const import card_ids, card_strings

"""
Utilities
"""


def get_cards_encoded(cards: List[int]) -> np.ndarray:
    """
    Get the 1-hot encoded array of the cards in the list.

    Args:
        cards: the cards

    Returns:
        1-hot encoded numpy array of the cards in the list
    """
    result = np.zeros(36, np.int32)
    result[cards] = 1
    return result


def get_cards_encoded_from_str(cards: List[str]) -> np.ndarray:
    """
    Get the 1-hot encoded array of the cards in the list.

    Args:
        cards: the cards

    Returns:
        1-hot encoded numpy array of the cards in the list
    """
    cards_int = convert_str_encoded_cards_to_int_encoded(cards)
    result = np.zeros(36, np.int32)
    result[cards_int] = 1
    return result


def convert_str_encoded_cards_to_int_encoded(cards: List[str]) -> List[int]:
    """
    Get the int encoded array of the str encoded cards in the list
    Args:
        cards: the cards as str encoded

    Returns:
        list of the cards, int encoded
    """
    return [card_ids[card] for card in cards]


def convert_int_encoded_cards_to_str_encoded(cards: List[int]) -> List[str]:
    """
    Get the int encoded array of the str encoded cards in the list
    Args:
        cards: the cards as int encoded

    Returns:
        list of the cards, str encoded
    """
    return [card_strings[i] for i in cards if i != -1]


def convert_one_hot_encoded_cards_to_str_encoded_list(cards: np.ndarray) -> List[str]:
    """
    Get the str encoded array of a one hot encoded array
    Args:
        cards: the cards, 1-hot encoded

    Returns:
        list of the cards as str
    """
    return [card_strings[i] for i in np.flatnonzero(cards)]


def convert_one_hot_encoded_cards_to_int_encoded_list(cards: np.ndarray) -> Iterable[int]:
    """
    Get the int encoded array of a one hot encoded array
    Args:
        cards: the cards, 1-hot encoded

    Returns:
        list of the cards as int
    """
    return np.flatnonzero(cards).tolist()


def count_colors(cards: np.ndarray) -> np.ndarray:
    """
    Count the colors in the cards. The return value is an array of size 4 that indicates how many cards of each
    color D, H, S and C are in the hand
    Args:
        cards: a one-hot encoded array of length 36 indicating the cards

    Returns:
        a an array of length for containing the number of cards of colors D, H, S and C
    """
    result = np.zeros(4, np.int32)
    cards.sum()
    result[0] = (cards[0:9]).sum()
    result[1] = (cards[9:18]).sum()
    result[2] = (cards[18:27]).sum()
    result[3] = (cards[27:36]).sum()
    return result


def deal_random_hand() -> np.ndarray:
    """
    Deal random cards for each hand.

    Returns:
        one hot encoded 4x36 array
    """
    # shuffle card ids
    cards = np.arange(0, 36, dtype=np.int32)
    np.random.shuffle(cards)
    hands = np.zeros(shape=[4, 36], dtype=np.int32)

    # convert to one hot encoded
    hands[0, cards[0:9]] = 1
    hands[1, cards[9:18]] = 1
    hands[2, cards[18:27]] = 1
    hands[3, cards[27:39]] = 1

    return hands






# HSLU
#
# Created by Thomas Koller on 7/24/2020
#
import numpy as np

from jass.game.game_observation import GameObservation
from jass.game.game_state import GameState


class GameRule:
    """
    Class for implementing rules of the jass match. The class includes rules that depend not on the process (like
    how trump is determined), but only upon cards. Currently this includes to determine the valid cards to play
    in a trick, to determine the winner of a trick and the points of a trick.

    This in an abstract base class that defines the interface.
    """

    def get_valid_cards(self, hand: np.array,
                        current_trick: np.ndarray or list,
                        move_nr: int,
                        trump: int or None) -> np.array:
        """
        Get the valid cards that can be played by the current player.

        Args:
            hand: one-hot encoded array of hands owned by the player
            current_trick: array with the indices of the cards for the previous moves in the current trick
            move_nr: which move the player has to make in the current trick, 0 for first move, 1 for second and so on
            trump: trump color (if used by the rule)

        Returns:
            one-hot encoded array of valid moves
        """
        raise NotImplementedError()

    def get_valid_cards_from_state(self, state: GameState):
        """
        Get the valid cards from the state for the current player.
        Args:
            state: The current state of the match
        Returns:
            one-hot encoded array of valid cards to play
        """
        return self.get_valid_cards(state.hands[state.player, :],
                                    state.current_trick,
                                    state.nr_cards_in_trick,
                                    state.trump)

    def get_valid_cards_from_obs(self, obs: GameObservation):
        """
        Get the valid cards from the observation.

        Precondition:
            The observation must be from the player whose turn it is too play (obs.player == obs.player_view)
        Args:
            obs: Observation of the match from players point of view
        Returns:
            one-hot encoded array of valid cards to play
        """
        return self.get_valid_cards(obs.hand,
                                    obs.current_trick,
                                    obs.nr_cards_in_trick,
                                    obs.trump)

    def calc_points(self, trick: np.ndarray, is_last: bool, trump: int = -1) -> int:
        """
        Calculate the points from the cards in the trick. Must be implemented in subclass

        Args:
            trick: the trick
            is_last: true if this is the last trick
            trump: the trump for the round (if needed by the rules)
        """
        raise NotImplementedError

    def calc_winner(self, trick: np.ndarray, first_player: int, trump: int = -1) -> int:
        """
        Calculate the winner of a completed trick. Must be implemented in subclass.

        Precondition:
            0 <= trick[i] <= 35, for i = 0..3
        Args:
            trick: the completed trick
            first_player: the first player of the trick
            trump: the trump for the round (if needed by the rules)

        Returns:
            the player who won this trick
        """
        raise NotImplementedError

    def assert_invariants(self, state: GameState) -> None:
        """
        Validates the internal consistency of the state according to the rules and throws an assertion exception if an
        error is detected.
        """
        raise NotImplementedError




# HSLU
#
# Created by Thomas Koller on 7/24/2020
#
import logging

import numpy as np

from jass.game.const import JASS_SCHIEBER, next_player, partner_player, card_ids
from jass.game.game_util import convert_int_encoded_cards_to_str_encoded, \
    convert_one_hot_encoded_cards_to_str_encoded_list, convert_str_encoded_cards_to_int_encoded


class GameObservation:
    """
    Observation of the state of the match from a player's view. This is the same as the GameState, except that
    a player does not see the hands (cards) of the other players:
        - the dealer
        - the player that declared trump,
        - the trump chosen
        - trump was declared forehand (derived information)
        - the tricks that have been played so far
        - the winner and the first player (derived) of each trick
        - the number of points that have been made by the current jass_players team in this round
        - the number of points that have been made by the opponent team in this round
        - the number of cards played in the current trick
        - the cards played in the current trick
        - the current player
        - the hand of the player

     Similarly to match state, the GameObservation captures the information in the following stages of the match:
    - Cards have been dealt, but no trump is selected yet (and it is the player's turn to select trump)
    - The first player that is allowed to choose trump has passed this right to the partner, and it is the
    partners player's turn to select trump
    - Trump has been declared by either player from the team that declares trump, but no card has been played yet
    - Between 1 and 35 cards have been played

    The observation is usually from the point of view of the current player, but can also be from the point of view
    of another players. In that case, the hand array are the cards of that player. In order to distinguish between
    these cases, the variable GameObservation.player_view is used.
    """

    # version of match observation (in json)
    FORMAT_VERSION = 'V0.2'

    def __init__(self) -> None:
        """
        Initialize the class. All numpy arrays will be allocated.
        """
        # dealer of the match
        self.dealer: int = -1

        # player of the next action, i.e. declaring trump or playing a card. T
        self.player: int = -1

        # player for which this observation is
        self.player_view = -1

        # selected trump
        self.trump: int = -1

        # true (1) if trump was declared forehand, 0 if it was declared rearhand, -1 if it has not been declared yet
        self.forehand: int = -1

        # the player, who declared trump (derived)
        self.declared_trump: int = -1

        #
        # information about held and played cards
        #

        # the hand cards of the player, 1-hot encoded
        self.hand = np.zeros(shape=36, dtype=np.int32)

        # the tricks played so far, with the cards of the tricks int encoded in the order they are played
        # a value of -1 indicates that the card has not been played yet
        self.tricks = np.full(shape=[9, 4], fill_value=-1, dtype=np.int32)

        # the winner of the tricks
        self.trick_winner = np.full(shape=9, fill_value=-1, dtype=np.int32)

        # the points made in the tricks
        self.trick_points = np.zeros(shape=9, dtype=np.int32)

        # the first player of the trick (derived)
        self.trick_first_player = np.full(shape=9, fill_value=-1, dtype=np.int32)

        # the current trick is a view onto self.trick
        self.current_trick = self.tricks[0, :]

        # the number of completed tricks
        self.nr_tricks = 0

        # the number of card in the current trick
        self.nr_cards_in_trick = 0

        # the total number of played cards (derived)
        self.nr_played_cards = 0

        self.points = np.zeros(shape=2, dtype=np.int32)

    def __eq__(self, other: 'GameObservation') -> bool:
        if self.nr_played_cards == 36:
            assert self.current_trick is None
            assert other.current_trick is None
            current_tricks_same = True
        else:
            current_tricks_same = (self.current_trick == other.current_trick).all()
        return \
            self.dealer == other.dealer and \
            self.player == other.player and \
            self.player_view == other.player_view and \
            self.trump == other.trump and \
            self.forehand == other.forehand and \
            self.declared_trump == other.declared_trump and \
            (self.hand == other.hand).all() and \
            (self.tricks == other.tricks).all() and \
            (self.trick_first_player == other.trick_first_player).all() and \
            (self.trick_winner == other.trick_winner).all() and \
            (self.trick_points == other.trick_points).all() and \
            self.nr_tricks == other.nr_tricks and \
            current_tricks_same and \
            self.nr_cards_in_trick == other.nr_cards_in_trick and \
            self.nr_played_cards == other.nr_played_cards and \
            (self.points == other.points).all()

    def __repr__(self):
        return str(self.__dict__)

    def to_json(self):
        """
        Generate a dict representation that can be converted to json. The format is the same than used in
        GameState.

        Returns:
            dict representation of the state
        """
        data = dict()

        # change of version: trump and player are always written, even as -1 if they are not known yet
        data['version'] = GameObservation.FORMAT_VERSION
        data['trump'] = int(self.trump)
        data['dealer'] = int(self.dealer)
        data['currentPlayer'] = int(self.player)
        data['playerView'] = int(self.player_view)
        data['forehand'] = int(self.forehand)

        # played tricks
        tricks = []

        # tricks
        for i in range(min(self.nr_tricks + 1, 9)):
            # add at most one more trick as finished (for first player)
            trick = {}
            # add cards (if not empty)
            cards = convert_int_encoded_cards_to_str_encoded(self.tricks[i, :].tolist())
            if len(cards) > 0:
                trick['cards'] = cards
                if len(cards) == 4:
                    # only add points and winners for complete tricks
                    trick['points'] = int(self.trick_points[i])
                    trick['win'] = int(self.trick_winner[i])

            if self.trick_first_player[i] != -1:
                trick['first'] = int(self.trick_first_player[i])
            if trick:
                # add if not empty dict
                tricks.append(trick)

        data['tricks'] = tricks

        # cards in the hand of the players (empty if no more cards)
        hand_empty = dict(hand=[])
        player_data = [hand_empty, hand_empty, hand_empty, hand_empty]
        hand = dict(hand=convert_one_hot_encoded_cards_to_str_encoded_list(self.hand))
        player_data[self.player_view] = hand
        data['player'] = player_data

        data['jassTyp'] = JASS_SCHIEBER
        return data

    @classmethod
    def from_json(cls, data: dict):
        """
        Create observation from dict (read from json)
        """
        obs = GameObservation()

        # version must be present for observation data
        if 'version' not in data:
            logging.getLogger(__name__).error('no version information')
            return None

        if data['version'] != GameObservation.FORMAT_VERSION:
            logging.getLogger(__name__).error('Unexpected format version: {}'.format(data['version']))
            return None

        # not currently used
        # jass_typ = data['jassTyp']

        obs.dealer = data['dealer']
        obs.player = data['currentPlayer']
        obs.player_view = data['playerView']
        obs.trump = data['trump']
        obs.forehand = data['forehand']

        if obs.trump != -1:
            if obs.forehand == 1:
                obs.declared_trump = next_player[obs.dealer]
            else:
                obs.declared_trump = partner_player[next_player[obs.dealer]]

        tricks = data['tricks']
        for i, trick in enumerate(tricks):
            # list of strings in the trick
            if 'cards' in trick:
                cards = trick['cards']
                obs.nr_played_cards += len(cards)

                # convert to list of ids
                cards_in_trick = convert_str_encoded_cards_to_int_encoded(cards)
                # append -1 if trick is not complete
                while len(cards_in_trick) < 4:
                    cards_in_trick.append(-1)

                obs.tricks[i] = np.array(cards_in_trick)
            if 'win' in trick:
                obs.trick_winner[i] = trick['win']
            if 'points' in trick:
                obs.trick_points[i] = trick['points']
            # first must be present for all tricks
            if 'first' in trick:
                obs.trick_first_player[i] = trick['first']
            else:
                logging.getLogger(__name__).error('No first player set in trick {}'.format(i))

        obs.nr_tricks, obs.nr_cards_in_trick = divmod(obs.nr_played_cards, 4)

        # current trick points to the correct trick
        if obs.nr_played_cards != 36:
            obs.current_trick = obs.tricks[obs.nr_tricks]
        else:
            obs.current_trick = None

        for i, player_data in enumerate(data['player']):
            if 'hand' in player_data and len(player_data['hand']) > 0:
                if i != obs.player_view:
                    logging.getLogger(__name__).error('Hand data for wrong player'.format(i))
                hand = player_data['hand']
                for card_constant in hand:
                    obs.hand[card_ids[card_constant]] = 1

        obs.points[0] = 0
        obs.points[1] = 0
        for trick in range(obs.nr_tricks):
            if obs.trick_winner[trick] == 0 or obs.trick_winner[trick] == 2:
                obs.points[0] += obs.trick_points[trick]
            else:
                obs.points[1] += obs.trick_points[trick]
        return obs




# HSLU
#
# Created by Thomas Koller on 03.08.20
#
import logging
import json
import numpy as np


class LogEntryFileGenerator:
    """
    Create a file of log entries with each entry in a separate line of the file. Files are split to contain
    no more than the indicated max_entries lines. Entries are first collected into a buffer and (optionally) shuffled
    before writing.

    The entries are the dict that have be generated using the appropriate serializer for the classes.

    The class should be used as a context manager within "with" in python.

    """
    EXTENSION = '.txt'

    def __init__(self, basename: str, max_entries: int, max_buffer: int=10000, shuffle=True):
        """
        Initialize the generator.

        Args:
            basename: basename of the generated files, this should include the whole file path
            max_entries: maximal entries per file
            max_buffer: size of the buffer that will be shuffled
        """
        self._basename = basename
        self._extension = LogEntryFileGenerator.EXTENSION
        self._max_entries = max_entries
        self._max_buffer = max_buffer
        self._current_file_number = 0
        self._nr_lines_in_file = 0
        self._shuffle = shuffle

        self._file = None
        self._buffer = []

        self._rng = np.random.default_rng()

    def __enter__(self):
        """
        Start of context region.
        """

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        End of context region, writes the buffer and closes the file.
        """
        if len(self._buffer) > 0:
            self._write_buffer()
        if self._file is not None:
            self._file.close()

    def _open_new_file(self):
        if self._file is not None:
            self._file.close()
        self._current_file_number += 1
        filename = self._basename + '{:04d}'.format(self._current_file_number) + self._extension
        logging.getLogger(__name__).info('Writing file: {}'.format(filename))
        self._file = open(filename, mode='w')
        self._nr_lines_in_file = 0

    def _write_buffer(self):
        if self._shuffle:
            self._rng.shuffle(self._buffer)
        for line in self._buffer:
            if self._nr_lines_in_file >= self._max_entries or self._file is None:
                self._open_new_file()

            self._file.write(line)
            self._file.write('\n')
            self._nr_lines_in_file += 1
        self._buffer.clear()

    def add_entry_line(self, line: str) -> None:
        """
        Add an entry line to the file. Adds the (already json formatted) line.
        Args:
            line: the line to add
        """
        self._buffer.append(line)
        if len(self._buffer) > self._max_buffer:
            self._write_buffer()

    def add_entry(self, entry: dict) -> None:
        """
        Add an entry to the file by converting it to a str first. The method is needed to be able to control
        the formatting of the json. So dicts should be added using this method, while lines that already
        contain a suitable representation can be added using add_entry_line.
        Args:
            entry: entry to add
        """
        line = json.dumps(entry, separators=(',', ':'))
        self.add_entry_line(line)



# HSLU
#
# Created by Thomas Koller on 12.10.18
#

from flask import Flask
from jass.agents.agent import Agent
from jass.service.player_service_route import players


class PlayerServiceApp(Flask):
    def __init__(self, import_name,
                 static_url_path=None,
                 static_folder='static',
                 static_host=None, host_matching=False,
                 subdomain_matching=False,
                 template_folder='templates',
                 instance_path=None,
                 instance_relative_config=False,
                 root_path=None):
        super(PlayerServiceApp, self).__init__(import_name,
                                               static_url_path=static_url_path,
                                               static_folder=static_folder,
                                               static_host=static_host,
                                               host_matching=host_matching,
                                               subdomain_matching=subdomain_matching,
                                               template_folder=template_folder,
                                               instance_path=instance_path,
                                               instance_relative_config=instance_relative_config,
                                               root_path=root_path)
        self.players = {}
        self.register_blueprint(players)

    def add_player(self, player_name: str, player: Agent):
        self.players[player_name] = player

    def get_player_for_name(self, player_name: str):
        return self.players[player_name] if player_name in self.players else None

    def get_players(self):
        return [name for name in self.players.keys()]



# HSLU
#
# Created by Thomas Koller on 7/31/2020
#
from datetime import datetime
from typing import List

from jass.game.const import DATE_FORMAT
from jass.game.game_state import GameState


class GameLogEntry:
    """
    Class to capture the information contained in match logs that are saved from real and simulated games
    and contain the match, the date and information about the players.
    """
    def __init__(self, game: GameState, date: datetime = None, player_ids: List[int] = None):
        """
        Game entry
        Args:
            game: a match state of a completed match
            date: the data when the match was played
            player_ids: information about the players
        """
        self.game = game
        self.date = date
        self.player_ids = player_ids

    def __eq__(self, other):
        return self.game == other.match and self.date == other.date and self.player_ids == other.player_ids

    def to_json(self) -> dict:
        """
        Convert to dict.

        Returns:
            dict containing the data
        """
        return dict(game=self.game.to_json(),
                    date=datetime.strftime(self.date, DATE_FORMAT),
                    player_ids=self.player_ids)

    @classmethod
    def from_json(cls, data):
        return GameLogEntry(game=GameState.from_json(data['match']),
                            date=datetime.strptime(data['date'], DATE_FORMAT),
                            player_ids=data['player_ids'])



# HSLU
#
# Created by Thomas Koller on 27.07.20
#
import numpy as np

class DealingCardStrategy:
    """
    Abstract base class to deal for a match in the arena

    """
    def deal_cards(self, game_nr: int=0, total_nr_games=0) -> np.ndarray:
        """
        Args:
            game_nr: number of games played so far, starting with 1 for the first match
            total_nr_games: total number of games to be played, or 0 if not used

        """
        raise NotImplementedError



# HSLU
#
# Created by Thomas Koller on 19.07.19
#
import logging

import numpy as np

from jass.game.game_util import convert_one_hot_encoded_cards_to_str_encoded_list, get_cards_encoded_from_str


class LabelPlay:
    """
    Class to define (possible) training information for a specific action in the match when it is in the playing
    stage (i.e. not in the trump defining stage).

    This includes the card played, the points made in the current trick by the own team and the other, the player who
    won the trick, the points made in the round by the own team and the other and the hands the players had at the
    beginning of the match.

    (Adding both the points for own and opposite team for the points in the trick eliminates needing to know the
    current player. Adding the information about the exact winner (instead of the team) might help to forecast the
    result of the trick)

    """
    def __init__(self,
                 card_played: int,
                 points_in_trick_own: int,
                 points_in_trick_other: int,
                 trick_winner: int,
                 points_in_round_own: int,
                 points_in_round_other: int,
                 hands: np.ndarray):
        self.card_played = card_played
        self.points_in_trick_own = points_in_trick_own
        self.points_in_trick_other = points_in_trick_other
        self.trick_winner = trick_winner
        self.points_in_round_own = points_in_round_own
        self.points_in_round_other = points_in_round_other
        self.hands = hands
        
    def to_json(self) -> dict: 
        """
        Convert to json
        Returns:
            dict that can be serialized to json
        """
        return dict(
            card_played=int(self.card_played),
            points_in_trick_own=int(self.points_in_trick_own),
            points_in_trick_other=int(self.points_in_trick_other),
            trick_winner=int(self.trick_winner),
            points_in_round_own=int(self.points_in_round_own),
            points_in_round_other=int(self.points_in_round_other),
            hands_player_0=convert_one_hot_encoded_cards_to_str_encoded_list(self.hands[0, :]),
            hands_player_1=convert_one_hot_encoded_cards_to_str_encoded_list(self.hands[1, :]),
            hands_player_2=convert_one_hot_encoded_cards_to_str_encoded_list(self.hands[2, :]),
            hands_player_3=convert_one_hot_encoded_cards_to_str_encoded_list(self.hands[3, :])
        )

    @classmethod
    def from_json(cls, data):
        """
        Create label from dict (from json)
        Args:
            data: dict representation
        Returns:
            label from the data
        """
        hands = np.zeros(shape=[4, 36], dtype=np.int32)
        try:
            hands[0, :] = get_cards_encoded_from_str(data['hands_player_0'])
            hands[1, :] = get_cards_encoded_from_str(data['hands_player_1'])
            hands[2, :] = get_cards_encoded_from_str(data['hands_player_2'])
            hands[3, :] = get_cards_encoded_from_str(data['hands_player_3'])
        except KeyError as e:
            logging.getLogger(__name__).error('Key error: {}, data: {}'.format(e, data))
            raise e

        return LabelPlay(card_played=data['card_played'],
                         points_in_trick_own=data['points_in_trick_own'],
                         points_in_trick_other=data['points_in_trick_other'],
                         trick_winner=data['trick_winner'],
                         points_in_round_own=data['points_in_round_own'],
                         points_in_round_other=data['points_in_round_other'],
                         hands=hands)




# HSLU
#
# Created by Thomas Koller on 7/24/2020
#
import numpy as np

from jass.game.const import color_of_card, color_masks, J_offset, higher_trump, lower_trump, card_values, UNE_UFE, \
    OBE_ABE, next_player, partner_player
from jass.game.game_rule import GameRule
from jass.game.game_state import GameState


class RuleSchieber(GameRule):
    """
    Class for implementing rules of the jass match for the variation for 'Schieber'. These are:
    - The match is played with trump

    """

    def get_valid_cards(self, hand: np.array,
                        current_trick: np.ndarray or list,
                        move_nr: int,
                        trump: int or None) -> np.array:
        """
        Get the valid cards that can be played by the current player. (It is implemented as one long function
        in order to take advantage of intermediate results of calculation.)

        Args:
            hand: one-hot encoded array of hands owned by the player
            current_trick: array with the indices of the cards for the previous moves in the current trick
            move_nr: which move the player has to make in the current trick, 0 for first move, 1 for second and so on
            trump: trump color (or 'obe', 'une')

        Returns:
            one-hot encoded array of valid moves
        """
        # play anything on the first move
        if move_nr == 0:
            return hand

        # get the color of the first played card and check if we have that color
        color_played = color_of_card[current_trick[0]]
        have_color_played = (np.sum(hand * color_masks[color_played, :]) > 0)

        if trump >= 4:
            # obe or une declared
            if have_color_played:
                # must give the correct color
                return hand * color_masks[color_played, :]
            else:
                # play anything, if we don't have the color
                return hand
        else:
            #
            # round with trumps declared (not 'obe' or 'une')
            #

            # check number of trumps we have and number of cards left, in order to simplify some of the conditions
            number_of_trumps = np.sum(hand * color_masks[trump, :])
            number_of_cards = np.sum(hand)

            #
            # the played color was trump
            #
            if color_played == trump:
                if number_of_trumps == 0:
                    # no more trumps, play anything
                    return hand
                elif number_of_trumps == 1:
                    if hand[trump * 9 + J_offset]:
                        # we have only trump jack, so we can play anything
                        return hand
                    else:
                        # we have just one trump and must play it
                        return hand * color_masks[trump, :]
                else:
                    # we have more than one trump, so we must play one of them
                    return hand * color_masks[trump, :]
            #
            # the played color was not trump
            #
            else:
                # check if anybody else (player 1 or player 2) played a trump, and if yes how high
                lowest_trump_played = None
                trump_played = False

                if move_nr > 1:
                    # check player 1
                    if color_of_card[current_trick[1]] == trump:
                        trump_played = True
                        lowest_trump_played = current_trick[1]
                    # check player 2
                    if move_nr == 3:
                        if color_of_card[current_trick[2]] == trump:
                            trump_played = True
                            if lowest_trump_played is not None:
                                # 2 trumps were played, so we must compare
                                if lowest_trump_played < current_trick[2]:
                                    # move from player 2 is lower (as its index value is higher)
                                    lowest_trump_played = current_trick[2]
                            else:
                                # this is the only trump played, so it is the lowest
                                lowest_trump_played = current_trick[2]

                #
                # nobody played a trump, so we do not need to consider any restrictions on playing trump ourselves
                #
                if not trump_played:
                    if have_color_played:
                        # must give a color or can give any trump
                        color_cards = hand * color_masks[color_played, :]
                        trump_cards = hand * color_masks[trump, :]
                        return color_cards + trump_cards
                    else:
                        # we do not have the color, so we can play anything, including any trump
                        return hand

                #
                # somebody played a trump, so we have the restriction that we can not play a lower trump, with
                # the exception if we only have trump left
                #
                else:
                    if number_of_trumps == number_of_cards:
                        # we have only trump left, so we can give any of them
                        return hand
                    else:
                        #
                        # find the valid trumps to play
                        #

                        # all trumps in hand
                        trump_cards = hand * color_masks[trump, :]

                        # higher trump cards in hand
                        higher_trump_cards = trump_cards * higher_trump[lowest_trump_played, :]

                        # lower trump cards in hand
                        lower_trump_cards = trump_cards * lower_trump[lowest_trump_played, :]

                    if have_color_played:
                        # must give a color or a higher trump
                        color_cards = hand * color_masks[color_played, :]
                        return color_cards + higher_trump_cards
                    else:
                        # play anything except a lower trump
                        not_lower_trump_cards = 1 - lower_trump_cards
                        return hand * not_lower_trump_cards



    def calc_points(self, trick: np.ndarray, is_last: bool, trump: int = -1) -> int:
        """
        Calculate the points from the cards in the trick according to the given trump

        Args:
            trick: the trick
            is_last: true if this is the last trick
            trump: trump for the round
        """
        return int(np.sum(card_values[trump, trick])) + (5 if is_last else 0)

    def calc_winner(self, trick: np.ndarray, first_player: int, trump: int = -1) -> int:
        """
        Calculate the winner of a completed trick.

        Second implementation in an attempt to be more efficient, while the implementation is somewhat longer
        and more complicated it is about 3 times faster than the previous method.

        Precondition:
            0 <= trick[i] <= 35, for i = 0..3
        Args:
            trick: the completed trick
            first_player: the first player of the trick
            trump: trump for the round
        Returns:
            the player who won this trick
        """
        color_of_first_card = color_of_card[trick[0]]
        if trump == UNE_UFE:
            # lowest card of first color wins
            winner = 0
            lowest_card = trick[0]
            for i in range(1, 4):
                # (lower card values have a higher card index)
                if color_of_card[trick[i]] == color_of_first_card and trick[i] > lowest_card:
                    lowest_card = trick[i]
                    winner = i
        elif trump == OBE_ABE:
            # highest card of first color wins
            winner = 0
            highest_card = trick[0]
            for i in range(1, 4):
                if color_of_card[trick[i]] == color_of_first_card and trick[i] < highest_card:
                    highest_card = trick[i]
                    winner = i
        elif color_of_first_card == trump:
            # trump mode and first card is trump: highest trump wins
            winner = 0
            highest_card = trick[0]
            for i in range(1, 4):
                # lower_trump[i,j] checks if j is a lower trump than i
                if color_of_card[trick[i]] == trump and lower_trump[trick[i], highest_card]:
                    highest_card = trick[i]
                    winner = i
        else:
            # trump mode, but different color played on first move, so we have to check for higher cards until
            # a trump is played, and then for the highest trump
            winner = 0
            highest_card = trick[0]
            trump_played = False
            trump_card = None
            for i in range(1, 4):
                if color_of_card[trick[i]] == trump:
                    if trump_played:
                        # second trump, check if it is higher
                        if lower_trump[trick[i], trump_card]:
                            winner = i
                            trump_card = trick[i]
                    else:
                        # first trump played
                        trump_played = True
                        trump_card = trick[i]
                        winner = i
                elif trump_played:
                    # color played is not trump, but trump has been played, so ignore this card
                    pass
                elif color_of_card[trick[i]] == color_of_first_card:
                    # trump has not been played and this is the same color as the first card played
                    # so check if it is higher
                    if trick[i] < highest_card:
                        highest_card = trick[i]
                        winner = i
        # adjust actual winner by first player
        return (first_player - winner) % 4

    def assert_invariants(self, state: GameState) -> None:
        """
        Validates the internal consistency of the state according to the rules and throws an assertion exception if an
        error is detected.
        """
        # trump declaration should be present
        if state.forehand is not None:
            if state.forehand:
                assert state.declared_trump == next_player[state.dealer]

            else:
                # either trump has been declared or not yet
                assert state.trump is None or state.declared_trump == partner_player[next_player[state.dealer]]
            assert state.dealer is not None

        # trick winners
        if state.nr_played_cards > 0:
            assert state.trick_first_player[0] == next_player[state.dealer]
        for i in range(1, state.nr_tricks):
            assert state.trick_winner[i - 1] == state.trick_first_player[i]

        # cards played
        assert state.nr_played_cards == 4 * state.nr_tricks + state.nr_cards_in_trick

        # total number of cards
        played_cards = state.tricks.flatten() > -1
        nr_played_cards = played_cards.sum()
        nr_cards_in_hand = state.hands.flatten().sum()
        assert nr_played_cards + nr_cards_in_hand == 36

        # number of points
        points_team_0 = 0
        points_team_1 = 0
        for trick in range(state.nr_tricks):
            if state.trick_winner[trick] == 0 or state.trick_winner[trick] == 2:
                points_team_0 += state.trick_points[trick]
            else:
                points_team_1 += state.trick_points[trick]
        assert points_team_0 == state.points[0]
        assert points_team_1 == state.points[1]

        # check current trick
        if state.nr_played_cards == 36:
            assert state.current_trick is None
        else:
            nr_cards_in_current_trick = np.count_nonzero(state.current_trick[:] > -1)
            expected_cards_in_current_trick = (state.nr_played_cards % 4)
            assert nr_cards_in_current_trick == expected_cards_in_current_trick




# HSLU
#
# Created by Thomas Koller on 05.11.19
#
import numpy as np

from jass.arena.dealing_card_strategy import DealingCardStrategy
from jass.game.game_util import deal_random_hand


class DealingCardRandomStrategy(DealingCardStrategy):
    """
    Deal cards randomly. This is the default implementation.
    """
    def deal_cards(self, game_nr: int=0, total_nr_games=0) -> np.ndarray:
        return deal_random_hand()



# HSLU
#
# Created by Ruedi Arnold on 16.01.2018
# Changed to use blueprint, Thomas Koller on 12.10.2018
# Changed to refactored jass-kit library
#

"""
Code for the Jass player web interface, i.e. the "web part" receiving requests and serving them accordingly.
This file handles requests like action_trump and action_play_card and delegates them to a one of the registered Jass
agents.
"""

import logging
from http import HTTPStatus

from flask import request, jsonify, Blueprint, current_app

from jass.game.const import card_strings
from jass.game.game_observation import GameObservation

JASS_PATH_PREFIX = '/jass/players/'
SELECT_TRUMP_PATH_PREFIX = '/action_trump'
PLAY_CARD_PATH_PREFIX = '/action_play_card'
SEND_INFO_PREFIX = '/game_info'

players = Blueprint(JASS_PATH_PREFIX, __name__)


@players.route('/<string:player_name>' + PLAY_CARD_PATH_PREFIX, methods=['POST'])
def action_play_card(player_name: str):
    """
    Takes a action_play_card request, validates its data and returns the card to play.
    Args:
        player_name: the name of the desired player
    Returns:
        the http response to answer the given request
    """
    # check if player registered at that name
    player = current_app.get_player_for_name(player_name)
    if player is None:
        logging.warning('player {} not found'.format(player_name))
        return jsonify(error='player not found'), HTTPStatus.BAD_REQUEST

    # check request type and parse data into a match observation
    if not request.is_json:
        logging.warning('request is not json')
        return jsonify(error='json data expected'), HTTPStatus.UNSUPPORTED_MEDIA_TYPE

    try:
        request_dict = request.get_json()
        obs = GameObservation.from_json(request_dict)
    except Exception as e:
        logging.warning('Error parsing request to GameObservation')
        return jsonify(error=str(e)), HTTPStatus.BAD_REQUEST

    try:
        card = player.action_play_card(obs)
        # convert card from int to string
        data = dict(card=card_strings[card])
        return jsonify(data), HTTPStatus.OK
    except Exception as e:
        logging.warning('Error in action_play_card')
        return jsonify(error=str(e)), HTTPStatus.INTERNAL_SERVER_ERROR


@players.route('/<string:player_name>' + SELECT_TRUMP_PATH_PREFIX, methods=['POST'])
def action_trump(player_name: str):
    """
    Takes a action_trump request, validates its data and returns the selected trump
    Args:
        player_name: the name of the desired player
    Returns:
        the http response to answer the given request

    """
    # check if player registered at that name

    player = current_app.get_player_for_name(player_name)
    if player is None:
        logging.warning('player {} not found'.format(player_name))
        return jsonify(error='player not found'), HTTPStatus.BAD_REQUEST

    # check request type and parse data into a match observation
    if not request.is_json:
        logging.warning('request is not json')
        return jsonify(error='json data expected'), HTTPStatus.UNSUPPORTED_MEDIA_TYPE

    try:
        request_dict = request.get_json()
        obs = GameObservation.from_json(request_dict)
    except Exception as e:
        logging.warning('Error parsing request to GameObservation')
        return jsonify(error=str(e)), HTTPStatus.BAD_REQUEST

    try:
        trump = player.action_trump(obs)
        data = dict(trump=trump)
        return jsonify(data), HTTPStatus.OK
    except Exception as e:
        logging.warning('Error in action_trump: {}'.format(e))
        return jsonify(error=str(e)), HTTPStatus.INTERNAL_SERVER_ERROR


@players.route('/<string:player_name>' + SEND_INFO_PREFIX, methods=['POST'])
def game_info(player_name: str):
    """
    Receives a match info message about a current changes in the match, which does not require
    an action from the player.

    This might be used to inform a player about cards played by the other player or the result
    at the end of the match.

    Args:
        player_name: the name of the desired player
    Returns:
        the http response to answer the given request

    """
    # check if player registered at that name
    player = current_app.get_player_for_name(player_name)
    if player is None:
        logging.warning('player {} not found'.format(player_name))
        return jsonify(error='player not found'), HTTPStatus.BAD_REQUEST

    # check request type and parse data into a match observation
    if not request.is_json:
        logging.warning('request is not json')
        return jsonify(error='json data expected'), HTTPStatus.UNSUPPORTED_MEDIA_TYPE

    try:
        request_dict = request.get_json()
        _ = GameObservation.from_json(request_dict)
        # currently we dont do anything with the data
        return jsonify(''), HTTPStatus.OK
    except Exception as e:
        logging.warning('Error parsing request to GameObservation')
        return jsonify(error=str(e)), HTTPStatus.BAD_REQUEST


@players.route('/<string:player_name>', methods=['GET'])
def check_player(player_name: str):
    """
    Provides basic information about this player.

    Args:
        player_name:  the player name as provided in the request path.

    Returns:
        different strings depending on the player found
    """
    player = current_app.get_player_for_name(player_name)
    if player is not None:
        return jsonify('Player {} here'.format(player_name)), HTTPStatus.OK
    else:
        return jsonify('No such player {}'.format(player_name)), HTTPStatus.NOT_FOUND



# HSLU
#
# Created by Thomas Koller on 24.07.18
#
"""
Constants, tables and small utility function for the jass match.
"""

# Implementation notes:
# - numpy arrays are used for all arrays
# - integer arrays are of type np.int32


import numpy as np

# format to read / write dates in dict/json
DATE_FORMAT = '%d.%m.%y %H:%M:%S'

#
# Numerical values correspond to the values in the obtained interface definition
#

# Players (players are numbered clockwise), playing is done counterclockwise
# noinspection PyPep8
NORTH       = 0                         # type: int
EAST        = 1                         # type: int
SOUTH       = 2                         # type: int
WEST        = 3                         # type: int
MAX_PLAYER  = 3                         # type: int

player_strings = ['North', 'East', 'South', 'West']

# Colors for trump, there is a long and a short constant for each
DIAMONDS    = 0                         # type: int  # Ecken / Schellen
D           = DIAMONDS
HEARTS      = 1                         # type: int  # Herz / Rosen
H           = HEARTS
SPADES      = 2                         # type: int  # Schaufeln / Schilten
S           = SPADES
CLUBS       = 3                         # type: int  # Kreuz / Eichel
C           = CLUBS
OBE_ABE     = 4                         # type: int
O           = OBE_ABE
UNE_UFE     = 5                         # type: int
U           = UNE_UFE
MAX_TRUMP   = 5                         # maximal value of a trump action (for loops)

# additional action available at trump selection phase, this is encoded as 10 in the Swisslos specification.
PUSH        = 10                        # type: int # Schieben
P           = PUSH

# alternative value for push (used in some ml approaches to be able to 1-hot encode the trump/push action
# (is translated to PUSH when the action is added to the round)
PUSH_ALT    = 6                         # type: int

# Strings for trumps
trump_strings_short = [
    'D', 'H', 'S', 'C', 'O', 'U', '', '', '', '', 'P'
]

# German strings for trumps
trump_strings_german_long = [
    'D: Schellen', 'H: Rosen', 'S: Schilten', 'C: Eichel', 'O: Obe-Abe', 'U: Une-Ufe'
]

# German string for 'Schieben'
trump_string_push_german = 'P: Schieben'

trump_ints = [
    DIAMONDS, HEARTS, SPADES, CLUBS, OBE_ABE, UNE_UFE
]

# Jass variants (jassTyp in request message)
JASS_SCHIEBER = 'SCHIEBER'
JASS_SCHIEBER_1000 = 'SCHIEBER_1000'
JASS_SCHIEBER_2500 = 'SCHIEBER_2500'
JASS_HEARTS = 'HEARTS'
JASS_ALL = [JASS_SCHIEBER_1000, JASS_SCHIEBER_2500, JASS_HEARTS]

# Card distributions (sets) like hands, cards played etc. will be modelled using arrays, corresponding to a 1-hot
# encoding. Also, this format will be used for neuronal networks. The following constants of all cards define the
# index of the cards in the array

# Diamonds
DA  = 0
DK  = 1
DQ  = 2
DJ  = 3
D10 = 4
D9  = 5
D8  = 6
D7  = 7
D6  = 8

# Hearts
HA  = 9
HK  = 10
HQ  = 11
HJ  = 12
H10 = 13
H9  = 14
H8  = 15
H7  = 16
H6  = 17

# Spades
SA  = 18
SK  = 19
SQ  = 20
SJ  = 21
S10 = 22
S9  = 23
S8  = 24
S7  = 25
S6  = 26

# Clubs
CA  = 27
CK  = 28
CQ  = 29
CJ  = 30
C10 = 31
C9  = 32
C8  = 33
C7  = 34
C6  = 35

# offsets from start of color
A_offset = 0
K_offset = 1
Q_offset = 2
J_offset = 3
Ten_offset = 4
Nine_offset = 5
Eight_offset = 6
Seven_offset = 7
Six_offset = 8

# offsets to start of cards for this color
color_offset = np.array([
    0, 9, 18, 27
], np.int32)

# array to generate string representations from the ids
card_strings = np.array([
    'DA',
    'DK',
    'DQ',
    'DJ',
    'D10',
    'D9',
    'D8',
    'D7',
    'D6',
    'HA',
    'HK',
    'HQ',
    'HJ',
    'H10',
    'H9',
    'H8',
    'H7',
    'H6',
    'SA',
    'SK',
    'SQ',
    'SJ',
    'S10',
    'S9',
    'S8',
    'S7',
    'S6',
    'CA',
    'CK',
    'CQ',
    'CJ',
    'C10',
    'C9',
    'C8',
    'C7',
    'C6',
], np.str)

# dictionary to get the ids from the strings:
# noinspection PyPep8
card_ids = {
    'DA' : DA ,
    'DK' : DK ,
    'DQ' : DQ ,
    'DJ' : DJ ,
    'D10': D10,
    'D9' : D9 ,
    'D8' : D8 ,
    'D7' : D7 ,
    'D6' : D6 ,
    'HA' : HA ,
    'HK' : HK ,
    'HQ' : HQ ,
    'HJ' : HJ ,
    'H10': H10,
    'H9' : H9 ,
    'H8' : H8 ,
    'H7' : H7 ,
    'H6' : H6 ,
    'SA' : SA ,
    'SK' : SK ,
    'SQ' : SQ ,
    'SJ' : SJ ,
    'S10': S10,
    'S9' : S9 ,
    'S8' : S8 ,
    'S7' : S7 ,
    'S6' : S6 ,
    'CA' : CA ,
    'CK' : CK ,
    'CQ' : CQ ,
    'CJ' : CJ ,
    'C10': C10,
    'C9' : C9 ,
    'C8' : C8 ,
    'C7' : C7 ,
    'C6' : C6
}
#
# 2D array of scoring values for the cards in a trick, each row is the scoring for the cards for the trump indicated
# by the row
#
# noinspection PyPep8
card_values = np.array(
    [
       # DA DK DQ DJ D10 D9 D8 D7 D6 HA HK HQ HJ H10 H9 H8 H7 H6 SA SK SQ SJ S10 S9 S8 S7 S6 CA CK CQ CJ C10 C9 C8 C7 C6
        [11, 4, 3,20, 10,14, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0],
        [11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3,20, 10,14, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0],
        [11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3,20, 10,14, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0],
        [11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3, 2, 10, 0, 0, 0, 0,11, 4, 3,20, 10,14, 0, 0, 0],
        [11, 4, 3, 2, 10, 0, 8, 0, 0,11, 4, 3, 2, 10, 0, 8, 0, 0,11, 4, 3, 2, 10, 0, 8, 0, 0,11, 4, 3, 2, 10, 0, 8, 0, 0],
        [ 0, 4, 3, 2, 10, 0, 8, 0,11, 0, 4, 3, 2, 10, 0, 8, 0,11, 0, 4, 3, 2, 10, 0, 8, 0,11, 0, 4, 3, 2, 10, 0, 8, 0,11]
    ], np.int32)
#
# 2D array of masks which cards belong to a specific color, can be used to filter matching/non-matching cards
#
color_masks = np.array(
    [
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    ], np.int32
)

#
# 1D array of which color a card belongs to
#
color_of_card = np.array(
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3],
    np.int32
)

#
# 1D array of offset of each card
offset_of_card = np.array(
    [A_offset, K_offset, Q_offset, J_offset, Ten_offset, Nine_offset, Eight_offset, Seven_offset, Six_offset,
     A_offset, K_offset, Q_offset, J_offset, Ten_offset, Nine_offset, Eight_offset, Seven_offset, Six_offset,
     A_offset, K_offset, Q_offset, J_offset, Ten_offset, Nine_offset, Eight_offset, Seven_offset, Six_offset,
     A_offset, K_offset, Q_offset, J_offset, Ten_offset, Nine_offset, Eight_offset, Seven_offset, Six_offset],
    np.int32
)
#
#
# 2D array of higher trump cards, i.e. higher_trump[DA,:] will give the mask of higher trumps than DA (i.e. DJ, DN)
# (used to make the logic for determining the valid moves slightly easier to read). The full array is constructed
# from a smaller array higher_trump_card for the card values only
#
higher_trump_card = np.array(
    [
        [0, 0, 0, 1, 0, 1, 0, 0, 0],  # A
        [1, 0, 0, 1, 0, 1, 0, 0, 0],  # K
        [1, 1, 0, 1, 0, 1, 0, 0, 0],  # Q
        [0, 0, 0, 0, 0, 0, 0, 0, 0],  # J
        [1, 1, 1, 1, 0, 1, 0, 0, 0],  # 10
        [0, 0, 0, 1, 0, 0, 0, 0, 0],  # 9
        [1, 1, 1, 1, 1, 1, 0, 0, 0],  # 8
        [1, 1, 1, 1, 1, 1, 1, 0, 0],  # 7
        [1, 1, 1, 1, 1, 1, 1, 1, 0]   # 6
    ], np.int32
)
# lower trump cards also include the card itself
lower_trump_card = 1 - higher_trump_card

# full array for each card
higher_trump = np.zeros([36, 36], np.int32)
higher_trump[0:9, 0:9] = higher_trump_card
higher_trump[9:18, 9:18] = higher_trump_card
higher_trump[18:27, 18:27] = higher_trump_card
higher_trump[27:36, 27:36] = higher_trump_card

# same for lower trump
lower_trump = np.zeros([36, 36], np.int32)
lower_trump[0:9, 0:9] = lower_trump_card
lower_trump[9:18, 9:18] = lower_trump_card
lower_trump[18:27, 18:27] = lower_trump_card
lower_trump[27:36, 27:36] = lower_trump_card

# next player of player with given index
next_player = [3, 0, 1, 2]

# partner player of player with given index
partner_player = [2, 3, 0, 1]

# true if same_player[i, j] are in the same team for players i, j
same_team = np.array(
    [
        [True, False, True, False],
        [False, True, False, True],
        [True, False, True, False],
        [False, True, False, True]
    ], np.bool)



# HSLU
#
# Created by Thomas Koller on 7/31/2020
#
from datetime import datetime

from jass.game.const import DATE_FORMAT
from jass.game.game_observation import GameObservation


class GameObsActionLogEntry:
    """
    Write logs containing the match observation for a single player and the action. The action could either be
    a trump action or a play card action (or combined).

    Date and player_id information is added so that entries could be filtered by player.
    """
    def __init__(self, obs: GameObservation, action: int, date: datetime, player_id: int):
        self.obs = obs
        self.action = action
        self.date = date
        self.player_id = player_id

    def __eq__(self, other: 'GameObsActionLogEntry'):
        return \
            self.obs == other.obs and \
            self.action == other.action and \
            self.date == other.date and \
            self.player_id == other.player_id

    def to_json(self) -> dict:
        """
        Convert to dict (for json)
        Returns:
            dict representation
        """
        return dict(obs=self.obs.to_json(),
                    action=self.action,
                    date=datetime.strftime(self.date, DATE_FORMAT),
                    player_id=self.player_id)

    @classmethod
    def from_json(cls, data: dict) -> 'GameObsActionLogEntry':
        """
        Convert data from dict to GameObsActionLogEntry
        Args:
            data: dict containing the serialized GameObsActionLogEntry
        Returns:
            GameObsActionLogEntry
        """
        return GameObsActionLogEntry(obs = GameObservation.from_json(data['obs']),
                                     action=int(data['action']),
                                     date=datetime.strptime(data['date'], DATE_FORMAT),
                                     player_id=int(data['player_id']))



# HSLU
#
# Created by Thomas Koller on 24.08.18
#
from jass.game.game_state import GameState

from jass.game.const import NORTH, EAST, SOUTH, WEST


class JassMatch:
    """
    A match consists of a number of (full) games that are played between 4 players.

    (Most of the logic and the available data is based on GameSim. The match is primarily used to store and load
    games that are played on the server. For that reason it is now extended to support the urls of the players
    and error that occurred while playing.
    """

    def __init__(self):
        self._players = ['', '', '', '']                # type: [str]
        self._urls = ['', '', '', '']                   # type: [str]
        self._player_ids = ['', '', '', '']             # type: [str]
        self._games = []                                # type: [GameState]
        self._points_team0 = 0                          # type: int
        self._points_team1 = 0                          # type: int
        self._winner = -1                               # type: int

        # time is not set by default, as the match can also be created by reading from db
        self._time_started = None
        self._time_finished = None

        # errors will be a list of str messages about possible errors while playing a match on the server
        # such as timeout or connection problems
        self._errors = []

    def __eq__(self, other: 'JassMatch'):
        """
        Compare two instances. Useful for tests when the representations are encoded and decoded. The objects are
        considered equal if they have the same properties. As the games contain numpy arrays, we can not compare
        dict directly.
        Args:
            other: the other object to compare to.

        Returns:
            True if the objects are the same.
        """
        result = \
            self._players == other._players and \
            self.points_team0 == other.points_team0 and \
            self.points_team1 == other.points_team1 and \
            self.time_started == other.time_started and \
            self.time_finished == other.time_finished and \
            self.errors == other.errors

        if not result:
            return False

        for i in range(self.nr_games):
            if not self._games[i] == other._games[i]:
                return False

        return True

    # Read only properties
    @property
    def points_team0(self) -> int:
        return self._points_team0

    @property
    def points_team1(self) -> int:
        return self._points_team1

    @property
    def round(self) -> [GameState]:
        return self._games

    @property
    def errors(self) -> [str]:
        return self._errors

    # Derived properties
    @property
    def nr_games(self) -> int:
        return len(self._games)

    # Read / Write properties
    @property
    def north(self) -> str:
        return self._players[NORTH]

    @north.setter
    def north(self, player: str):
        self._players[NORTH] = player

    @property
    def east(self) -> str:
        return self._players[EAST]

    @east.setter
    def east(self, player: str):
        self._players[EAST] = player

    @property
    def south(self) -> str:
        return self._players[SOUTH]

    @south.setter
    def south(self, player: str):
        self._players[SOUTH] = player

    @property
    def west(self) -> str:
        return self._players[WEST]

    @west.setter
    def west(self, player: str):
        self._players[WEST] = player

    @property
    def north_url(self) -> str:
        return self._urls[NORTH]

    @north_url.setter
    def north_url(self, url: str):
        self._urls[NORTH] = url

    @property
    def east_url(self) -> str:
        return self._urls[EAST]

    @east_url.setter
    def east_url(self, url: str):
        self._urls[EAST] = url

    @property
    def south_url(self) -> str:
        return self._urls[SOUTH]

    @south_url.setter
    def south_url(self, url: str):
        self._urls[SOUTH] = url

    @property
    def west_url(self) -> str:
        return self._urls[WEST]

    @west_url.setter
    def west_url(self, url: str):
        self._urls[WEST] = url

    @property
    def north_id(self):
        return self._player_ids[NORTH]

    @property
    def east_id(self):
        return self._player_ids[EAST]

    @property
    def south_id(self):
        return self._player_ids[SOUTH]

    @property
    def west_id(self):
        return self._player_ids[WEST]

    @property
    def winner(self) -> int:
        return self._winner

    @winner.setter
    def winner(self, value: int):
        self._winner = value

    @property
    def time_started(self):
        return self._time_started

    @time_started.setter
    def time_started(self, value):
        self._time_started = value

    @property
    def time_finished(self):
        return self._time_finished

    @time_finished.setter
    def time_finished(self, value):
        self._time_finished = value

    def set_players(self, north: str, east: str, south: str, west: str) -> None:
        """
        Set the players.
        Args:
            north: North player
            east: East player
            south: South player
            west: West player
        """
        self._players[NORTH] = north
        self._players[EAST] = east
        self._players[SOUTH] = south
        self._players[WEST] = west

    def set_urls(self, north_url: str, east_url: str, south_url: str, west_url: str) -> None:
        """
        Set the players.
        Args:
           north_url: url of north player
           east_url: url of east player
           south_url: url of south player
           west_url: url of west player
        """
        self._urls[NORTH] = north_url
        self._urls[EAST] = east_url
        self._urls[SOUTH] = south_url
        self._urls[WEST] = west_url

    def set_player_ids(self, north_id: str, east_id: str, south_id: str, west_id: str) -> None:
        """
        Set the players.
        Args:
           north_id: id of north player
           east_id: id of east player
           south_id: id of south player
           west_id: id of west player
        """
        self._player_ids[NORTH] = north_id
        self._player_ids[EAST] = east_id
        self._player_ids[SOUTH] = south_id
        self._player_ids[WEST] = west_id

    def add_game(self, game: GameState) -> None:
        """
        Add a match to the match. The points are adjusted from the match.
        Args:
            game: state of completed match to add

        """
        self._games.append(game)
        self._points_team0 += int(game.points[0])
        self._points_team1 += int(game.points[1])

    def add_error(self, error: str) -> None:
        """
        Add an error message.
        Args:
            error: Error message to add
        """
        self._errors.append(error)

    def to_json(self):
        """
        Generate a dict representation that can be converted to json.


        Returns:
            dict representation of the state
        """

        data = dict()
        data['north'] = self.north
        data['east'] = self.east
        data['south'] = self.south
        data['west'] = self.west
        data['northUrl'] = self.north_url
        data['eastUrl'] = self.east_url
        data['southUrl'] = self.south_url
        data['westUrl'] = self.west_url
        data['northId'] = self.north_id
        data['eastId'] = self.east_id
        data['southId'] = self.south_id
        data['westId'] = self.west_id
        data['winner'] = self.winner
        data['pointsTeam0'] = int(self.points_team0)
        data['pointsTeam1'] = int(self.points_team1)
        data['timeStarted'] = self.time_started
        data['timeFinished'] = self.time_finished

        if self.errors:
            data['errors'] = self.errors

        games = []
        for i in range(self.nr_games):
            game_data = self._games[i].to_json()
            games.append(game_data)
        data['games'] = games
        return data

    @classmethod
    def from_json(cls, data: dict):
        """
        Create match from dict (read from json)
        """
        match = JassMatch()

        match.set_players(data['north'], data['east'], data['south'], data['west'])
        match.winner = data['winner']
        match._points_team0 = data['pointsTeam0']
        match._points_team1 = data['pointsTeam1']
        match.time_started = data['timeStarted']
        match.time_finished = data['timeFinished']

        if 'errors' in data:
            match._errors = data['errors']

        games = data['games']

        games_read = []

        for game_data in games:
            game = GameState.from_json(game_data)
            games_read.append(game)

        match._games = games_read
        return match




# HSLU
#
# Created by Thomas Koller on 7/22/2020
#

# Simulator for a match

import copy
import numpy as np

from jass.game.const import next_player, PUSH, partner_player, NORTH, SOUTH
from jass.game.game_rule import GameRule
from jass.game.game_state import GameState
from jass.game.game_state_util import observation_from_state


class GameSim:
    """
    Class for simulating a match. A match consists of selecting trump and playing 36 cards, currently the trump selection
    phase is implemented for Schieber. For other versions of the match regarding the order of actions, the class can be
    overridden. The actual rules of the match, points scoring, winning tricks and calculating which cards are allowed
    to be played are implemented in the rule class.
    """
    def __init__(self, rule: GameRule):
        # the internal state of the match is stored in a GameState object that can be set and retrieved
        self._state = GameState()
        self._rule = rule

    def init_from_state(self, state: GameState):
        self._state = copy.deepcopy(state)

    def init_from_cards(self, hands: np.array, dealer: int):
        self._state.dealer = dealer
        self._state.player = next_player[dealer]
        self._state.trump = -1
        self._state.forehand = -1
        self._state.hands = hands.copy()
        self._state.tricks.fill(-1)
        self._state.trick_winner.fill(-1)
        self._state.trick_points.fill(0)
        self._state.trick_first_player.fill(-1)
        self._state.current_trick = self._state.tricks[0, :]
        self._state.nr_tricks = 0
        self._state.nr_cards_in_trick = 0
        self._state.nr_played_cards = 0
        self._state.points[0] = 0
        self._state.points[1] = 0

    @property
    def rule(self):
        return self._rule

    @property
    def state(self):
        return self._state

    def get_observation(self):
        """
        Get the observation for the current player in the current state of the match.

        Returns:
            The observation for the current player.
        """
        return observation_from_state(self._state)

    def action_trump(self, action: int) -> None:
        if self._state.forehand == -1:
            # this is the action of the forehand player
            if action == PUSH:
                self._state.forehand = 0
                self._state.player = partner_player[self._state.player]
            else:
                self._state.forehand = 1
                self._state.trump = action
                self._state.declared_trump = self._state.player
                self._state.trick_first_player[0] = self._state.player
                # player remains the same
        elif self._state.forehand == 0:
            # action of the partner of the forehand player
            self._state.trump = action
            self._state.declared_trump = self._state.player
            self._state.player = next_player[self._state.dealer]
            self._state.trick_first_player[0] = self._state.player
        else:
            raise ValueError('Unexpected value {} for forehand in action_trump'.format(self._state.forehand))

    def action_play_card(self, card: int) -> None:
        """
        Play a card as the current player and update the state of the round.

        Preconditions:
            self._state.nr_played_cards < 36
            self._state.hands[self.player,card] == 1
            (trump selection done according to rules of the jass variation)

        Postconditions:
            see assert_invariants

        Args:
            card: The card to play
        """
        # remove card from player
        self._state.hands[self._state.player, card] = 0

        # place in trick
        self._state.current_trick[self._state.nr_cards_in_trick] = card
        self._state.nr_played_cards += 1

        if self._state.nr_cards_in_trick < 3:
            if self._state.nr_cards_in_trick == 0:
                # make sure the first player is set on the first card of a new trick
                # (it will not have been set, if the round has been restored from dict)
                self._state.trick_first_player[self._state.nr_tricks] = self._state.player
            # trick is not yet finished
            self._state.nr_cards_in_trick += 1
            self._state.player = next_player[self._state.player]
        else:
            # finish current trick
            self._end_trick()

    def is_done(self):
        """
        Return true if the match is finished.

        Returns:
            true if the match is finished, false otherwise
        """
        return self._state.nr_played_cards == 36

    def _end_trick(self) -> None:
        """
        End the current trick and update all the necessary fields.
        """
        # update information about the current trick
        points = self._rule.calc_points(self._state.current_trick, self._state.nr_played_cards == 36, self._state.trump)
        self._state.trick_points[self._state.nr_tricks] = points
        winner = self._rule.calc_winner(self._state.current_trick, self._state.trick_first_player[self._state.nr_tricks], self._state.trump)
        self._state.trick_winner[self._state.nr_tricks] = winner

        if winner == NORTH or winner == SOUTH:
            self._state.points[0] += points
        else:
            self._state.points[1] += points
        self._state.nr_tricks += 1
        self._state.nr_cards_in_trick = 0

        if self._state.nr_tricks < 9:
            # not end of round
            # next player is the winner of the trick
            self._state.trick_first_player[self._state.nr_tricks] = winner
            self._state.player = winner
            self._state.current_trick = self._state.tricks[self._state.nr_tricks, :]
        else:
            # end of round
            self._state.player = -1
            self._state.current_trick = None


# HSLU
#
# Created by Thomas Koller on 7/30/2020
#
import logging

import requests

from jass.agents.agent import Agent
from jass.agents.agent_random_schieber import AgentRandomSchieber
from jass.game.const import card_ids
from jass.game.game_observation import GameObservation
from jass.service.player_service_route import SEND_INFO_PREFIX, SELECT_TRUMP_PATH_PREFIX, PLAY_CARD_PATH_PREFIX


class AgentByNetwork(Agent):
    """
    Forwards the request to a player service. Used for locally playing against deployed services.

    A random agent is used as standing player, if the service does not answer within a timeout.
    """

    def __init__(self, url, timeout=10):
        self._logger = logging.getLogger(__name__)
        self._standin_player = AgentRandomSchieber()
        self._base_url = url
        self._url_info = self._base_url + SEND_INFO_PREFIX
        self._url_trump = self._base_url + SELECT_TRUMP_PATH_PREFIX
        self._url_play = self._base_url + PLAY_CARD_PATH_PREFIX
        self._timeout = timeout

    def action_trump(self, obs: GameObservation) -> int:
        data = obs.to_json()
        data['gameId'] = 0
        try:
            self._logger.info('Sending request...')
            response = requests.post(self._url_trump, json=data, timeout=self._timeout)
            response_data = response.json()
            self._logger.info('got response: {}'.format(response_data))
            trump = int(response_data['trump'])
            return trump
        except Exception:
            self._logger.error('No response from network player, using standin player')
            return self._standin_player.action_trump(obs)

    def action_play_card(self, obs: GameObservation) -> int:
        data = obs.to_json()
        data['gameId'] = 0
        try:
            self._logger.info('Sending request...')
            response = requests.post(self._url_play, json=data, timeout=self._timeout)
            response_data = response.json()
            self._logger.info('got response: {}'.format(response_data))
            card = response_data['card']
            card_id = card_ids[card]
            return card_id
        except Exception:
            self._logger.error('No response from network player, using standin player')
            return self._standin_player.action_play_card(obs)



# HSLU
#
# Created by Thomas Koller on 7/28/2020
#
from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jass_kit",
    version="2.0.1",
    author="ABIZ HSLU",
    author_email="thomas.koller@hslu.ch",
    description="Package for the match of jass",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=find_namespace_packages(include=['jass.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy>=1.19'
    ],
    python_requires='>=3.6'
)




# HSLU
#
# Created by Thomas Koller on 27.07.20
#
import logging
import sys
from typing import List
import numpy as np

from jass.agents.agent import Agent
from jass.arena.dealing_card_random_strategy import DealingCardRandomStrategy
from jass.arena.dealing_card_strategy import DealingCardStrategy
from jass.game.const import NORTH, EAST, SOUTH, WEST, DIAMONDS, MAX_TRUMP, PUSH, next_player
from jass.game.game_observation import GameObservation
from jass.game.game_sim import GameSim
from jass.game.rule_schieber import RuleSchieber


class Arena:
    """
    Class for arenas. An arena plays a number of games between two pairs of players. The number of
    games to be played can be specified. The arena keeps statistics of the games won by each side and also
    of the point difference when winning.

    The class uses some strategy and template methods patterns. Currently, this is only done for dealing cards.

    """

    def __init__(self,
                 nr_games_to_play: int,
                 dealing_card_strategy: DealingCardStrategy = None,
                 print_every_x_games: int = 5,
                 check_move_validity=True,
                 save_filename=None):
        """

        Args:
            nr_games_to_play: number of games in the arena
            dealing_card_strategy: strategy for dealing cards
            print_every_x_games: print results every x games
            check_move_validity: True if moves from the agents should be checked for validity
            save_filename: True if results should be save
        """
        self._logger = logging.getLogger(__name__)

        self._nr_games_played = 0
        self._nr_games_to_play = nr_games_to_play

        # the strategies
        if dealing_card_strategy is None:
            self._dealing_card_strategy = DealingCardRandomStrategy()
        else:
            self._dealing_card_strategy = dealing_card_strategy

        # the players
        self._players: List[Agent or None] = [None, None, None, None]

        # player ids to use in saved games (if written)
        self._player_ids: List[int] = [0, 0, 0, 0]

        # the current match that is being played
        self._game = GameSim(rule=RuleSchieber())

        # we store the points for each match
        self._points_team_0 = np.zeros(self._nr_games_to_play)
        self._points_team_1 = np.zeros(self._nr_games_to_play)

        # Print  progress
        self._print_every_x_games = print_every_x_games
        self._check_moves_validity = check_move_validity

        # Save file if enabled
        if save_filename is not None:
            self._save_games = True
            #self._file_generator = RoundLogEntryFileGenerator(basename=save_filename, max_entries=100000)
        else:
            self._save_games = False

    @property
    def nr_games_to_play(self):
        return self._nr_games_to_play

        # We define properties for the individual players to set/get them easily by name

    @property
    def north(self) -> Agent:
        return self._players[NORTH]

    @north.setter
    def north(self, player: Agent):
        self._players[NORTH] = player

    @property
    def east(self) -> Agent:
        return self._players[EAST]

    @east.setter
    def east(self, player: Agent):
        self._players[EAST] = player

    @property
    def south(self) -> Agent:
        return self._players[SOUTH]

    @south.setter
    def south(self, player: Agent):
        self._players[SOUTH] = player

    @property
    def west(self) -> Agent:
        return self._players[WEST]

    @west.setter
    def west(self, player: Agent):
        self._players[WEST] = player

    @property
    def players(self):
        return self._players

    # properties for the results (no setters as the values are set by the strategies using the add_win_team_x methods)
    @property
    def nr_games_played(self):
        return self._nr_games_played

    @property
    def points_team_0(self):
        return self._points_team_0

    @property
    def points_team_1(self):
        return self._points_team_1

    def get_observation(self) -> GameObservation:
        """
        Creates and returns the observation for the current player

        Returns:
            the observation for the current player
        """
        return self._game.get_observation()

    def set_players(self, north: Agent, east: Agent, south: Agent, west: Agent,
                    north_id=0, east_id=0, south_id=0, west_id=0) -> None:
        """
        Set the players.
        Args:
            north: North player
            east: East player
            south: South player
            west: West player
            north_id: id to use for north in the save file
            east_id: id to use for east in the save file
            south_id: id to use for south in the save file
            west_id: id to use for west in the save file
        """
        self._players[NORTH] = north
        self._players[EAST] = east
        self._players[SOUTH] = south
        self._players[WEST] = west
        self._player_ids[NORTH] = north_id
        self._player_ids[EAST] = east_id
        self._player_ids[SOUTH] = south_id
        self._player_ids[WEST] = west_id

    def play_game(self, dealer: int) -> None:
        """
        Play a complete match (36 cards).
        """
        # init match
        self._game.init_from_cards(dealer=dealer, hands=self._dealing_card_strategy.deal_cards(
            game_nr=self._nr_games_played,
            total_nr_games=self._nr_games_to_play))

        # determine trump
        # ask first player
        trump_action = self._players[self._game.state.player].action_trump(self._game.get_observation())
        if trump_action < DIAMONDS or (trump_action > MAX_TRUMP and trump_action != PUSH):
            self._logger.error('Illegal trump (' + str(trump_action) + ') selected')
            raise RuntimeError('Illegal trump (' + str(trump_action) + ') selected')
        self._game.action_trump(trump_action)
        if trump_action == PUSH:
            # ask second player
            trump_action = self._players[self._game.state.player].action_trump(self._game.get_observation())
            if trump_action < DIAMONDS or trump_action > MAX_TRUMP:
                self._logger.error('Illegal trump (' + str(trump_action) + ') selected')
                raise RuntimeError('Illegal trump (' + str(trump_action) + ') selected')
            self._game.action_trump(trump_action)

        # play cards
        for cards in range(36):
            obs = self._game.get_observation()
            card_action = self._players[self._game.state.player].action_play_card(obs)
            if self._check_moves_validity:
                assert card_action in np.flatnonzero(self._game.rule.get_valid_cards_from_obs(obs)), \
                    'Invalid card played!'
            self._game.action_play_card(card_action)

        # update results
        self._points_team_0[self._nr_games_played] = self._game.state.points[0]
        self._points_team_1[self._nr_games_played] = self._game.state.points[1]
        self.save_game()

        self._nr_games_played += 1

    def save_game(self):
        """
        Save the current match if enabled.
        Returns:

        """
        #if self._save_games:
        #    round_log_entry = RoundLogEntry(rnd=self._rnd, date=datetime.now(), player_ids=self._player_ids)
        #    self._file_generator.add_entry(round_log_entry)

    def play_all_games(self):
        """
        Play the number of games.
        """
        #if self._save_games:
        #    self._file_generator.__enter__()
        dealer = NORTH
        for game_id in range(self._nr_games_to_play):
            self.play_game(dealer=dealer)
            if self.nr_games_played % self._print_every_x_games == 0:
                points_to_write = int(self.nr_games_played / self._nr_games_to_play * 40)
                spaces_to_write = 40 - points_to_write
                sys.stdout.write("\r[{}{}] {:4}/{:4} games played".format('.' * points_to_write,
                                                                          ' ' * spaces_to_write,
                                                                          self.nr_games_played,
                                                                          self._nr_games_to_play))
            dealer = next_player[dealer]
        #if self._save_games:
            # TODO: enable saving
            # self._file_generator.__exit__(None, None, None)
        sys.stdout.write('\n')




# HSLU
#
# Created by Thomas Koller on 7/24/2020
#
import numpy as np

from jass.game.const import next_player, partner_player
from jass.game.game_observation import GameObservation
from jass.game.game_state import GameState


def calculate_starting_hands_from_game(game: GameState) -> np.ndarray:
    """
    Calculate the hands of the players at the beginning of the match from a completed match.
    Args:
        game: state of a completed match

    Returns:
        array with hand cards of each player

    """
    hands = np.zeros(shape=[4, 36], dtype=np.int32)
    for rnd_nr in range(0, 9):
        player = game.trick_first_player[rnd_nr]
        for card_nr in range(4):
            card_played = game.tricks[rnd_nr, card_nr]
            hands[player, card_played] = 1
            player = next_player[player]
    return hands


def calculate_points_from_tricks(state: 'GameState') -> np.ndarray:
    """
    Calculate the points of the teams from the trick points and trick winners in the state.
    Args:
        state: state from which to calculate points

    Returns:
        array containing the points of the players
    """
    points = np.zeros(2, dtype=np.int32)
    for trick in range(state.nr_tricks):
        if state.trick_winner[trick] == 0 or state.trick_winner[trick] == 2:
            points[0] += state.trick_points[trick]
        else:
            points[1] += state.trick_points[trick]
    return points


def observation_from_state(state: GameState, player: int = -1) -> GameObservation:
    """
    Initialize observation from match state for the given player or the current player if the player is not
    supplied.

    Args:
        state: The match state from which to determine the observation
        player: player for which to create the observation or -1 for the current player

    Returns:
        the observation for a given match state for the view of the player
    """
    obs = GameObservation()

    obs.dealer = state.dealer
    obs.player = state.player

    if player == -1:
        obs.player_view = state.player
    else:
        obs.player_view = player

    obs.trump = state.trump
    obs.forehand = state.forehand
    obs.declared_trump = state.declared_trump

    if state.nr_played_cards < 36:
        obs.hand[:] = state.hands[obs.player_view, :]

    obs.tricks[:, :] = state.tricks[:, :]
    obs.trick_winner[:] = state.trick_winner[:]
    obs.trick_points[:] = state.trick_points[:]
    obs.trick_first_player[:] = state.trick_first_player[:]
    obs.nr_tricks = state.nr_tricks
    obs.nr_cards_in_trick = state.nr_cards_in_trick

    # current trick is a view to the trick
    if state.nr_played_cards < 36:
        obs.current_trick = obs.tricks[obs.nr_tricks]
    else:
        obs.current_trick = None
    obs.nr_tricks = state.nr_tricks
    obs.nr_cards_in_trick = state.nr_cards_in_trick
    obs.nr_played_cards = state.nr_played_cards
    obs.points[:] = state.points[:]

    return obs


def state_from_complete_game(game: GameState, cards_played: int) -> GameState:
    """
    Create the state of a match from the state of a completed match for a specific card played

    Preconditions:
        0 <= cards_played <= 35
        match.nr_played_cards == 36

    Args:
        game: The state of the completed match from which to create the state.
        cards_played: the number of cards played for which the state should be created

    Returns:
        a GameState object for the state when the cards have been played.

    """
    state = GameState()
    state.dealer = game.dealer
    state.trump = game.trump
    state.forehand = game.forehand
    state.declared_trump = game.declared_trump

    state.nr_played_cards = cards_played

    # calculate the number of tricks played and how many cards in the current trick
    state.nr_tricks, state.nr_cards_in_trick = divmod(cards_played, 4)

    # copy the trick first player, this is also available after making trump, when no trick has been played yet
    state.trick_first_player[0:state.nr_tricks + 1] = game.trick_first_player[0:state.nr_tricks + 1]

    if cards_played > 0:
        if state.nr_cards_in_trick == 0:
            # only full tricks
            state.tricks[0:state.nr_tricks, :] = game.tricks[0:state.nr_tricks, :]

            # current trick is empty (or none if last card)
            if cards_played == 36:
                state.current_trick = None
            else:
                # this is the next trick, after the full ones
                state.current_trick = state.tricks[state.nr_tricks]

        else:
            # copy all the full tricks first
            state.tricks[0:state.nr_tricks, :] = game.tricks[0:state.nr_tricks, :]

            # copy the trick in progress
            state.tricks[state.nr_tricks, 0:state.nr_cards_in_trick] = \
                game.tricks[state.nr_tricks, 0:state.nr_cards_in_trick]
            # make sure the current trick points to that
            state.current_trick = state.tricks[state.nr_tricks]
        # copy the results from the tricks
        state.trick_winner[0:state.nr_tricks] = game.trick_winner[0:state.nr_tricks]
        state.trick_points[0:state.nr_tricks] = game.trick_points[0:state.nr_tricks]

        state.points[:] = calculate_points_from_tricks(state)

        # determine player
        state.player = (game.trick_first_player[state.nr_tricks]-state.nr_cards_in_trick) % 4
    else:
        # no cards played yet
        state.player = next_player[state.dealer]

    # determine hand still held by the player, which are the cards that the player will play in the next
    # tricks of the full match, that are not played yet

    # add cards for completed tricks
    nr_tricks_completed = state.nr_tricks
    if state.nr_cards_in_trick != 0:
        # there is a non complete trick, treat it specially by adding the cards not yet played in the current trick
        player = (game.trick_first_player[state.nr_tricks]-state.nr_cards_in_trick) % 4
        for card_nr in range(state.nr_cards_in_trick, 4):
            card_played = game.tricks[state.nr_tricks, card_nr]
            state.hands[player, card_played] = 1
            player = next_player[player]
        # exclude this trick from the full tricks
        nr_tricks_completed += 1
    for game_nr in range(nr_tricks_completed, 9):
        player = game.trick_first_player[game_nr]
        for card_nr in range(4):
            card_played = game.tricks[game_nr, card_nr]
            state.hands[player, card_played] = 1
            player = next_player[player]
    return state


def state_for_trump_from_complete_game(game: GameState, for_forhand: bool) -> GameState:
    """
    Create the state of a match from the state of a completed match for the trump selection.

    Preconditions:
        0 <= cards_played <= 35
        match.nr_played_cards == 36

    Args:
        game: The state of the completed match from which to create the state.
        for_forhand: Create the state for the forehand player

    Returns:
        a GameState object for the state when the trump has been selected

    """
    state = GameState()

    state.dealer = game.dealer
    state.trump = -1

    if for_forhand:
        state.forehand = -1
        state.player = next_player[state.dealer]
    else:
        if game.forehand == 1:
            raise ValueError('Requested action for backhand, when match was forehand ')
        else:
            state.forehand = 0
            state.player = partner_player[next_player[state.dealer]]

    state.nr_played_cards = 0

    state.hands[:, :] = calculate_starting_hands_from_game(game)

    return state
